Framework BRL.Max2D
Import BRL.GLMax2D
Import BRL.Timer
Import Bah.Random
Import BRL.StandardIO
Import BRL.PNGLoader

SuperStrict

Graphics 1024, 768

Const MAPSIZEX:Int = 1024
Const MAPSIZEY:Int = 768

Const MAX_BOUNCE:Int = 10

Global MapData:Byte[,] = LoadVoxelData()
Global Normals:Int[,] = LoadNormalData()
Global MapLighting:Double[MAPSIZEX, MAPSIZEY, 3]

Local Tree:TVoxelQuadTree = New TVoxelQuadTree.Create(MapData)

Global Lightmap:TPixmap = CreatePixmap(MAPSIZEX, MAPSIZEY, PF_RGBA8888)

TLight.Create(TLight.KIND_LINE,   258.0,  82.0, 1.0, 100.0, 255, 255, 255)
TLight.Create(TLight.KIND_LINE,   760.0,  82.0, 1.0, 100.0, 255, 255, 255)
TLight.Create(TLight.KIND_LINE,   915.0, 450.0, 1.0, 100.0, 255, 255, 255).Angle = 90.0
TLight.Create(TLight.KIND_LINE,   915.0, 620.0, 1.0, 100.0, 255, 255, 255).Angle = 90.0
TLight.Create(TLight.KIND_LINE,   108.0, 536.0, 1.0, 100.0, 255, 255, 255).Angle = -90.0
TLight.Create(TLight.KIND_LINE,   413.0, 156.0, 1.0, 100.0, 255, 255, 255).Angle = -90.0
TLight.Create(TLight.KIND_SPHERE, 350.0, 540.0, 1.0,  50.0, 255, 128, 128)

Local Timer:TTimer = CreateTimer(30)
Local ElapsedTime:Int, ElapsedTimer:Int, TimeElapsed:Int
While Not (KeyHit(KEY_ESCAPE) Or AppTerminate())
	Cls
	
	ElapsedTimer:Int = MilliSecs()
	TLight.CalcLighting(Tree, 1000)
	TimeElapsed :+ MilliSecs() - ElapsedTimer
	RenderMap()
	
	DrawText "Raycount: " + TLight.TotalRayCount, 0, 0
	DrawText "Rays/s: " + TLight.TotalRayCount/(TimeElapsed*0.001), 0, 22
	
	Flip 0
	WaitTimer(Timer)
Wend

End

Type TLight
	Const KIND_LINE  :Int = 1
	Const KIND_SPHERE:Int = 2
	
	Global Lights:TList = New TList
	
	Global TotalRayCount:Int
	
	Field Kind:Int
	
	Field X:Double
	Field Y:Double
	Field Z:Double
	
	Field Angle:Double
	
	Field Size:Double
	Field Intensity:Double
	
	Field Color:Byte[3]
	
	Method New()
		Lights.AddLast(Self)
	End Method
	
	Method CalcSphereLighting(Tree:TVoxelQuadTree, RayCount:Int)
		Local RayIntensity:Double = Intensity
		
		For Local I:Int = 1 To RayCount
			Local Phi:Double = RndDouble()*360.0
			Local Theta:Double = RndDouble()*90.0
			
			Local Radius:Double = (RndDouble() - 0.5)*Size
			Local PosAngle:Double = RndDouble()*360.0
			Local RayX:Double = X + Radius*Cos(PosAngle)
			Local RayY:Double = Y + Radius*Sin(PosAngle)
			
			Local DirX:Double = Cos(Phi)*Cos(Theta)
			Local DirY:Double = Sin(Phi)*Cos(Theta)
			Local DirZ:Double = Sin(Theta)
			
			If Not TraceRay(Tree, RayX, RayY, 200.0, DirX, DirY, DirZ) Then Continue
			
			If RayX > 0.0 And RayY > 0.0 And RayX < MAPSIZEX - 1.0 And RayY < MAPSIZEY - 1.0 Then
				Local TileX:Int = Int(RayX)
				Local TileY:Int = Int(RayY)
				
				Local U:Double = RayX - TileX
				Local V:Double = RayY - TileY
				
				For Local I:Int = 0 To 2
					MapLighting[TileX,     TileY,     I] :+ Color[I]*RayIntensity*(1.0 - U)*(1.0 - V)
					MapLighting[TileX + 1, TileY,     I] :+ Color[I]*RayIntensity*       U *(1.0 - V)
					MapLighting[TileX + 1, TileY + 1, I] :+ Color[I]*RayIntensity*       U *       V
					MapLighting[TileX,     TileY + 1, I] :+ Color[I]*RayIntensity*(1.0 - U)*       V
				Next
			EndIf
		Next
		
		TotalRayCount :+ RayCount
	End Method
	
	Method CalcLineLighting(Tree:TVoxelQuadTree, RayCount:Int)
		Local RayIntensity:Double = Intensity
		
		For Local I:Int = 1 To RayCount
			Local Phi:Double = RndDouble()*180.0
			Local Theta:Double = ACos(Sqr(RndDouble()))
			
			Local Radius:Double = (RndDouble() - 0.5)*Size
			Local RayX:Double = X + Radius*Cos(Angle)
			Local RayY:Double = Y + Radius*Sin(Angle)
			
			Local DirX:Double = Cos(Phi)*Cos(Theta)
			Local DirY:Double = Sin(Theta)
			Local DirZ:Double = Sin(Phi)*Cos(Theta)
			
			Local NewDirX:Double = Cos(Angle)*DirX - Sin(Angle)*DirY
			Local NewDirY:Double = Sin(Angle)*DirX + Cos(Angle)*DirY
			
			DirX = NewDirX
			DirY = NewDirY
			
			If I < RayCount/4 Then
				If Not TraceRay(Tree, RayX, RayY, 4.0, DirX, DirY, DirZ) Then Continue
			Else
				If Not TraceRay(Tree, RayX, RayY, 40.0, DirX, DirY, DirZ) Then Continue
			EndIf
			
			If RayX > 0.0 And RayY > 0.0 And RayX < MAPSIZEX - 1.0 And RayY < MAPSIZEY - 1.0 Then
				Local TileX:Int = Int(RayX)
				Local TileY:Int = Int(RayY)
				
				Local U:Double = RayX - TileX
				Local V:Double = RayY - TileY
				
				For Local I:Int = 0 To 2
					MapLighting[TileX,     TileY,     I] :+ Color[I]*RayIntensity*(1.0 - U)*(1.0 - V)
					MapLighting[TileX + 1, TileY,     I] :+ Color[I]*RayIntensity*       U *(1.0 - V)
					MapLighting[TileX + 1, TileY + 1, I] :+ Color[I]*RayIntensity*       U *       V
					MapLighting[TileX,     TileY + 1, I] :+ Color[I]*RayIntensity*(1.0 - U)*       V
				Next
			EndIf
		Next
		
		TotalRayCount :+ RayCount
	End Method
	
	Method TraceRay:Int(Tree:TVoxelQuadTree, X:Double Var, Y:Double Var, Z:Double, DirX:Double, DirY:Double, DirZ:Double)
		Local Factor:Double = 1.0/Float(Max(MAPSIZEX, MAPSIZEY))
		Local NX:Double, NY:Double, Normal:Int, Iteration:Int
		While Z > 0.0 And Iteration < MAX_BOUNCE
			Local T:Double = Tree.Raymarch(X*Factor + 1.0, Y*Factor + 1.0, DirX, DirY, Normal)*1024.0
			
			If Z < DirZ*T Then
				X :+ DirX*Z/DirZ
				Y :+ DirY*Z/DirZ
				
				Z = 0.0
			Else
				Z :- DirZ*T
				X :+ DirX*(T - 0.01)
				Y :+ DirY*(T - 0.01)
				
				UncompressNormal(Normal, NX, NY)
				
				Local Phi:Double = RndDouble()*180.0
				Local Theta:Double = ACos(Sqr(RndDouble()))
				
				DirX = Cos(Phi)*Cos(Theta)
				DirY = Sin(Theta)
				DirZ = Sin(Phi)*Cos(Theta)
				
				Local Angle:Double = ATan2(NY, NX) - 90.0
				
				Local NewDirX:Double = Cos(Angle)*DirX - Sin(Angle)*DirY
				Local NewDirY:Double = Sin(Angle)*DirX + Cos(Angle)*DirY
				
				DirX = NewDirX
				DirY = NewDirY
				
				Iteration :+ 1
			EndIf
		Wend
		
		If Z <> 0.0 Then Return False Else Return True
	End Method
	
	Function CalcLighting(Tree:TVoxelQuadTree, RaysPerLight:Int)
		For Local Light:TLight = EachIn Lights
			Select Light.Kind
				Case KIND_SPHERE
					Light.CalcSphereLighting(Tree, RaysPerLight)
				Case KIND_LINE
					Light.CalcLineLighting(Tree, RaysPerLight)
			End Select
		Next
	End Function
	
	Function Create:TLight(Kind:Int, X:Double, Y:Double, Intensity:Double = 1.0, Size:Double = 20.0, R:Byte = 255, G:Byte = 255, B:Byte = 255)
		Local Light:TLight = New TLight
		
		Light.Kind = Kind
		Light.X = X
		Light.Y = Y
		Light.Size = Size
		Light.Intensity = Intensity
		Light.Color[0] = R
		Light.Color[1] = G
		Light.Color[2] = B
		
		Return Light
	End Function
End Type

Type TVoxelQuadTree
	Const MAX_SCALE:Int = 23
	Const EPSILON:Double = 2^-MAX_SCALE
	
	Global LeafPadding:Int[] = [ ..
		0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4, ..
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, ..
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, ..
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, ..
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, ..
		4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8 ..
	]
	
	
	Field Data:Int Ptr
	
	Method Raymarch:Double( OX:Double, OY:Double, DX:Double, DY:Double, Normal:Int Var )
		Local Stack:Int[ MAX_SCALE + 1, 2 ]
		
		If Abs( DX ) < EPSILON Then DX = EPSILON
		If Abs( DY ) < EPSILON Then DY = EPSILON
		
		Local dTx:Double = 1.0/( -Abs( DX ) )
		Local dTy:Double = 1.0/( -Abs( DY ) )
		
		Local bTx:Double = dTx*OX
		Local bTy:Double = dTy*OY
		
		Local OctantMask:Int = %11
		If DX > 0.0 Then OctantMask = OctantMask ~ %01; bTx = 3.0*dTx - bTx
		If DY > 0.0 Then OctantMask = OctantMask ~ %10; bTy = 3.0*dTy - bTy
		
		Local MinT:Double = Max( 2.0*dTx - bTx, 2.0*dTy - bTy )
		Local MaxT:Double = Min( 1.0*dTx - bTx, 1.0*dTy - bTy )
		Local H:Double = MaxT
		MinT = Max( MinT, 0.0 )
		'MaxT = Min( MaxT, 1.0 )
		
		Local Current:Int = 0
		Local Parent:Int  = 0
		Local IDx:Int     = 0
		Local PosX:Double  = 1.0
		Local PosY:Double  = 1.0
		Local Scale:Int   = MAX_SCALE - 1
		
		Local Scale_Exp2:Double = 0.5
		
		If 1.5*dTx - bTx > MinT Then IDx = IDx ~ %01; PosX = 1.5
		If 1.5*dTy - bTy > MinT Then IDx = IDx ~ %10; PosY = 1.5
		
		While Scale < MAX_SCALE
			If Current = 0 Then Current = Data[ Parent ]
			
			Local CornerTX:Double = PosX*dTx - bTx
			Local CornerTY:Double = PosY*dTy - bTy
			Local MaxTC:Double = Min( CornerTX, CornerTY )
			
			Local ChildShift:Int = IDx ~ OctantMask
			Local ChildMasks:Int = Current Shl ChildShift
			
			If ( ChildMasks & $8000 ) And MinT <= MaxT Then
				Local MaxTV:Double = Min( MaxT, MaxTC )
				Local Half:Double  = Scale_Exp2*0.5
				Local CenterTX:Double = Half*dTx + CornerTX
				Local CenterTY:Double = Half*dTy + CornerTY
				
				If MinT <= MaxTV Then
					Local ChildOffset:Int = Current Shr 16
					
					If ChildMasks & $0080 Then
						Normal = Data[ChildOffset + Parent + LeafPadding[((ChildMasks Shr (8 + ChildShift)) Shl ChildShift) & 127]]
						
						Exit
					EndIf
					
					If MaxTC < H Then
						Stack[ Scale, 0 ] = Parent
						Stack[ Scale, 1 ] = FloatToInt( MaxT )
					EndIf
					
					H = MaxTC
					
					ChildOffset :+ LeafPadding[ ( ~ChildMasks ) & %1110000 ]
					Parent :+ ChildOffset
					
					IDx = 0
					Scale :- 1
					Scale_Exp2 = Half
					
					If CenterTX > MinT Then IDx = IDx ~ %01; PosX :+ Scale_Exp2
					If CenterTY > MinT Then IDx = IDx ~ %10; PosY :+ Scale_Exp2
					
					MaxT = MaxTV
					Current = 0
					Continue
				EndIf
			EndIf
			
			Local StepMask:Int
			If CornerTX <= MaxTC Then StepMask = StepMask ~ %01; PosX :- Scale_Exp2
			If CornerTY <= MaxTC Then StepMask = StepMask ~ %10; PosY :- Scale_Exp2
			
			MinT = MaxTC
			IDx = IDx ~ StepMask
			
			If IDx & StepMask Then
				Local DifferingBits:Int
				
				If StepMask & %01 Then DifferingBits = DifferingBits | ( FloatToInt( PosX ) ~ FloatToInt( PosX + Scale_Exp2 ) )
				If StepMask & %10 Then DifferingBits = DifferingBits | ( FloatToInt( PosY ) ~ FloatToInt( PosY + Scale_Exp2 ) )
				Scale = ( FloatToInt( Float( DifferingBits ) ) Shr 23 ) - 127
				Scale_Exp2 = IntToFloat( ( Scale - MAX_SCALE + 127 ) Shl 23 )
				
				Parent = Stack[ Scale, 0 ]
				MaxT = IntToFloat( Stack[ Scale, 1 ] )
				
				Local ShX:Int = FloatToInt( PosX ) Shr Scale
				Local ShY:Int = FloatToInt( PosY ) Shr Scale
				PosX = IntToFloat( ShX Shl Scale )
				PosY = IntToFloat( ShY Shl Scale )
				IDx = ( ShX & 1 ) | ( ( ShY & 1 ) Shl 1 )
				
				H = 0.0
				Current = 0
			EndIf
		Wend
		
		If Scale >= MAX_SCALE Then Return 1000000.0
		
		If ( OctantMask & %01 ) = 0 Then PosX = 3.0 - Scale_Exp2 - PosX
		If ( OctantMask & %10 ) = 0 Then PosY = 3.0 - Scale_Exp2 - PosY
		
		Return MinT
	End Method
	
	Function DrawEmptyQuad( X:Double, Y:Double, S:Double )
		DrawLine X, Y, X + S, Y
		DrawLine X, Y, X, Y + S
		DrawLine X, Y + S, X + S, Y + S
		DrawLine X + S, Y, X + S, Y + S
	End Function
	
	Function FloatToInt:Int( In:Float )
		Return ( Int Ptr ( Varptr In ) )[ 0 ]
	End Function
	
	Function IntToFloat:Float( In:Int )
		Return ( Float Ptr ( Varptr In ) )[ 0 ]
	End Function
	
	Method QuadContainsVoxels:Int( Voxels:Byte[,], StartX:Int, StartY:Int, Size:Int )
		If Size = 1 Then Return Voxels[ StartX, StartY ] > 0
		
		For Local X:Int = StartX Until StartX + Size
			For Local Y:Int = StartY Until StartY + Size
				If Voxels[ X, Y ] = 255 Then Return True
			Next
		Next
	End Method
	
	Method BuildQuadTree:Int( Voxels:Byte[,], X:Int, Y:Int, Size:Int, DescriptorIndex:Int, WriteIndex:Int )
		Local HalfSize:Int = Size Shr 1
		
		Local Children:Byte, Leafs:Byte, ChildIndex:Int, ChildCount:Int
		
		If QuadContainsVoxels( Voxels, X + HalfSize, Y + HalfSize, HalfSize ) Then Children = Children | %10000000; ChildCount :+ 1
		If QuadContainsVoxels( Voxels, X           , Y + HalfSize, HalfSize ) Then Children = Children | %01000000; ChildCount :+ 1
		If QuadContainsVoxels( Voxels, X + HalfSize, Y           , HalfSize ) Then Children = Children | %00100000; ChildCount :+ 1
		If QuadContainsVoxels( Voxels, X           , Y           , HalfSize ) Then Children = Children | %00010000; ChildCount :+ 1
		
		Leafs = ~Children
		
		Local ChildWriteIndex:Int
		
		If HalfSize = 1 Then
			Leafs = %11110000
			
			ChildWriteIndex = WriteIndex - 1
			
			For Local I:Int = 0 Until 8
				If Children & (1 Shl I) Then
					ChildWriteIndex :+ 1
					
					Data[ChildWriteIndex] = Normals[X + (Sgn(I & 1))*HalfSize, Y + (Sgn(I & 2))*HalfSize]
				EndIf
			Next
		Else
			ChildWriteIndex = WriteIndex + ChildCount
			
			Local ChildrenWritten:Int
			For Local I:Int = 0 Until 4
				If Children & ( 16 Shl I ) Then
					ChildWriteIndex = BuildQuadTree( Voxels, X + ( Sgn( I & 1 ) )*HalfSize, Y + ( Sgn( I & 2 ) )*HalfSize, HalfSize, WriteIndex + ChildrenWritten, ChildWriteIndex ) + 1
					
					ChildrenWritten :+ 1
				EndIf
			Next
		EndIf
		
		ChildIndex = WriteIndex - DescriptorIndex
		
		If ChildIndex > 65535 Then RuntimeError "ABANDON SHIP"
		
		Data[ DescriptorIndex ] = ( ChildIndex Shl 16 ) | ( Children Shl 8 ) | Leafs
		
		Return ChildWriteIndex
	End Method
	
	Method Render( X:Int, Y:Int, Size:Int = 512, Index:Int = 0 )
		Local HalfSize:Int = Size Shr 1
		
		DrawLine X, Y, X + Size - 1, Y
		DrawLine X, Y + Size - 1, X + Size - 1, Y + Size - 1
		DrawLine X, Y + Size - 1, X, Y
		DrawLine X + Size - 1, Y + Size - 1, X + Size - 1, Y
		
		Local Descriptor:Int = Data[ Index ]
		Local ChildIndex:Int = Descriptor Shr 16
		Local Children:Int = ( Descriptor Shr 8 ) & $FF
		Local Leafs:Int = Descriptor & $FF
		
		Local ChildrenDrawn:Int
		For Local I:Int = 0 Until 4
			If Children & ( 16 Shl I ) And Not ( Leafs & ( 16 Shl I ) ) Then
				Render( X + ( Sgn( I & 1 ) )*HalfSize, Y + ( Sgn( I & 2 ) )*HalfSize, HalfSize, Index + ChildIndex + ChildrenDrawn )
				
				ChildrenDrawn :+ 1
			EndIf
		Next
	End Method
	
	Method Create:TVoxelQuadTree( Voxels:Byte[,] )
		If Data Then MemFree( Data )
		
		Data = Int Ptr MemAlloc( 1024*1024*4 )
		
		Print BuildQuadTree( Voxels, 0, 0, 1024, 0, 1 )
		
		Return Self
	End Method
End Type

Function RenderMap()
	Local MaxValue:Double = -1, MinValue:Double = 1000000000.0
	
	For Local Y:Int = 0 Until MAPSIZEY
		For Local X:Int = 0 Until MAPSIZEX
			MaxValue = Max(MaxValue, Max(MapLighting[X, Y, 2], Max(MapLighting[X, Y, 1], MapLighting[X, Y, 0])))
			MinValue = Min(MinValue, Min(MapLighting[X, Y, 2], Min(MapLighting[X, Y, 1], MapLighting[X, Y, 0])))
		Next
	Next
	
	For Local Y:Int = 0 Until MAPSIZEY
		For Local X:Int = 0 Until MAPSIZEX
			If MapData[X, Y] Then
				Lightmap.WritePixel(X, Y, $FFFFFF)
			Else
				Local ExpFactor:Double = 0.5 'Use cheapo exponential tone mapping
				Local R:Byte = Int(((MapLighting[X, Y, 0] - MinValue)/(MaxValue - MinValue))^ExpFactor*255.0)
				Local G:Byte = Int(((MapLighting[X, Y, 1] - MinValue)/(MaxValue - MinValue))^ExpFactor*255.0)
				Local B:Byte = Int(((MapLighting[X, Y, 2] - MinValue)/(MaxValue - MinValue))^ExpFactor*255.0)
				
				Lightmap.WritePixel(X, Y, $FF000000 | (R Shl 16) | (G Shl 8) | B)
			EndIf
		Next
	Next
	
	DrawPixmap Lightmap, 0, 0
	
End Function

Function CompressNormal:Int(NX:Double, NY:Double) 'Compresses a 2D normal into a 32bit integer
	Const USCALE:Int = (1 Shl 29) - 1
	
	If Abs(NX) > Abs(NY) Then
		Local Face:Int = 0
		Local Sign:Int = (NX < 0.0)
		
		Local U:Int = Int(((NY/Abs(NX))*0.5 + 0.5)*USCALE)
		
		Return (Face Shl 30) | (Sign Shl 29) | U
	Else
		Local Face:Int = 1
		Local Sign:Int = (NY < 0.0)
		
		Local U:Int = Int(((NX/Abs(NY))*0.5 + 0.5)*USCALE)
		
		Return (Face Shl 30) | (Sign Shl 29) | U
	EndIf
End Function

Function UncompressNormal(Normal:Int, NX:Double Var, NY:Double Var)
	Const USCALE:Double = 1.8626451527004039896659627364314e-9
	
	Local Face:Int = (Normal Shr 30) & 1
	Local Sign:Int = (Normal Shr 29) & 1
	Local U   :Int = Normal & $1FFFFFFF
	
	If Face = 1 Then
		NX = (U*USCALE - 0.5)*2.0
		NY = 1 - Sign*2
	Else
		NX = 1 - Sign*2
		NY = (U*USCALE - 0.5)*2.0
	EndIf
	
	Local Length:Double = Sqr(NX*NX + NY*NY)
	NX :/ Length
	NY :/ Length
End Function

Function LoadVoxelData:Byte[,]()
	Local Data:Byte[1024, 1024]
	Local VoxelData:Byte[1024, 1024]
	
	Local Pixmap:TPixmap = LoadPixmap("Office.png")
	
	For Local X:Int = 0 Until Pixmap.Width
		For Local Y:Int = 0 Until Pixmap.Height
			Data[X, Y] = Pixmap.ReadPixel(X, Y) & $FF
			If Data[X, Y] > 0 Then Data[X, Y] = 255
		Next
	Next
	
	For Local X:Int = 1 Until Pixmap.Width - 1
		For Local Y:Int = 1 Until Pixmap.Height - 1
			If Data[X, Y] = 255 And (Data[X - 1, Y] = 0 Or Data[X, Y - 1] = 0 Or Data[X + 1, Y] = 0 Or Data[X, Y + 1] = 0) Then
				VoxelData[X, Y] = 255
			Else
				VoxelData[X, Y] = 0
			EndIf
		Next
	Next
	
	Return VoxelData
End Function

Function LoadNormalData:Int[,]()
	Local Data:Int[1024, 1024]
	Local NormalData:Int[1024, 1024]
	
	Local Pixmap:TPixmap = LoadPixmap("Office.png")
	
	For Local X:Int = 0 Until Pixmap.Width
		For Local Y:Int = 0 Until Pixmap.Height
			Data[X, Y] = Pixmap.ReadPixel(X, Y) & $FF
			If Data[X, Y] > 0 Then Data[X, Y] = 255
		Next
	Next
	
	For Local X:Int = 1 Until Pixmap.Width - 1
		For Local Y:Int = 1 Until Pixmap.Height - 1
			If Data[X, Y] Then
				Local DX:Int = Sgn(Data[X - 1, Y] - Data[X + 1, Y])
				Local DY:Int = Sgn(Data[X, Y - 1] - Data[X, Y + 1])
				
				NormalData[X, Y] = CompressNormal(DX, DY)
			EndIf
		Next
	Next
	
	Return NormalData
End Function